from vmad import operator, autooperator
from pmesh.pm import ParticleMesh
from vmad.lib import linalg
import numpy

@operator
class to_scalar:
    ain = {'x' : 'RealField'}
    aout = {'y' : '*'}

    def apl(node, x):
        return dict(y = x.cnorm())

    def vjp(node, _y, x):
        return dict(_x= x * (2 * _y))

    def jvp(node, x_, x):
        return dict(y_ = x.cdot(x_) * 2)

@operator
class as_complex_field:
    ain = {'x' : '*'}
    aout = {'y' : 'ComplexField'}

    def apl(node, x, pm):
        y = pm.create(type='complex')
        y.real[...] = x[..., 0]
        y.imag[...] = x[..., 1]
        return dict(y=y)

    def vjp(node, _y, pm):
        _x = numpy.stack([_y.real, _y.imag], axis=-1)
        return dict(_x=_x)

    def jvp(node, x_, pm):
        y_ = pm.create(type='complex')
        y_.real[...] = x_[..., 0]
        y_.imag[...] = x_[..., 1]
        return dict(y_=y_)

@operator
class r2c:
    ain = {'x' : 'RealField'}
    aout = {'y' : 'ComplexField'}

    def apl(node, x):
        return dict(y=x.r2c(), pm=x.pm)
    def vjp(node, _y, pm):
        _y = pm.create(type='complex', value=_y)
        return dict(_x=_y.r2c_vjp())
    def jvp(node, x_, pm):
        x_ = pm.create(type='real', value=x_)
        return dict(y_=x_.r2c())

@operator
class c2r:
    ain = {'x' : 'ComplexField'}
    aout = {'y' : 'RealField'}

    def apl(node, x):
        return dict(y=x.c2r(), pm=x.pm)
    def vjp(node, _y, pm):
        _y = pm.create(type='real', value=_y)
        return dict(_x=_y.c2r_vjp())
    def jvp(node, x_, pm):
        x_ = pm.create(type='complex', value=x_)
        return dict(y_=x_.c2r())

@operator
class apply_transfer:
    ain = {'x' : 'ComplexField'}
    aout = {'y' : 'ComplexField'}

    def apl(node, x, tf):
        filter = lambda k, v: v * tf(k)
        return dict(y=x.apply(filter))

    def vjp(node, _y, tf):
        filter = lambda k, v: v * numpy.conj(tf(k))
        return dict(_x=_y.apply(filter))

    def jvp(node, x_, tf):
        filter = lambda k, v: v * tf(k)
        return dict(y_=x_.apply(filter))

@operator
class paint:
    aout = {'mesh' : 'RealField'}
    ain =  {'x': 'ndarray',
           'layout': 'Layout',
           'mass' : 'ndarray'
           }

    def apl(node, x, mass, layout, pm):
        mesh = pm.paint(x, mass=mass, layout=layout, hold=False)
        return dict(mesh=mesh)

    def vjp(node, _mesh, x, mass, layout, pm):
        _mesh = pm.create(type='real', value=_mesh)
        _x, _mass = pm.paint_vjp(_mesh, x, layout=layout, mass=mass)
        return dict(
            _layout = 0,
            _x=_x,
            _mass=_mass)

    def jvp(node, x_, x, layout, mass, layout_, mass_, pm):
        if x_ is 0: x_ = None
        if mass_ is 0: mass_ = None # force cast it to a scalar 0, so make it None
        mesh_ = pm.paint_jvp(x, v_mass=mass_, mass=mass, v_pos=x_, layout=layout)

        return dict(mesh_=mesh_)

@operator
class readout:
    aout = {'value' : 'ndarray'}

    ain = {'x': 'ndarray',
        'mesh': 'RealField',
      'layout' : 'Layout'}

    def apl(node, mesh, x, layout, resampler=None):
        N = mesh.pm.comm.allreduce(len(x))
        value = mesh.readout(x, layout=layout, resampler=resampler)
        return dict(value=value, pm=mesh.pm)

    def vjp(node, _value, x, layout, mesh, pm, resampler=None):
        _mesh, _x = mesh.readout_vjp(x, _value, layout=layout, resampler=resampler)
        return dict(_mesh=_mesh, _x=_x, _layout=0)

    def jvp(node, x_, mesh_, x, layout, layout_, mesh, pm, resampler=None):
        if mesh_ is 0: mesh_ = None
        if x_ is 0: x_ = None
        mesh = pm.create(type='real', value=mesh)
        value_ = mesh.readout_jvp(x, v_self=mesh_, v_pos=x_, layout=layout, resampler=resampler)
        return dict(value_=value_)

@operator
class decompose:
    aout={'layout' : 'Layout'}
    ain={'x': 'ndarray'}

    def apl(node, x, pm):
        return dict(layout=pm.decompose(x))

    def vjp(engine, _layout):
        return dict(_x=0)

    def jvp(engine, x_):
        return dict(layout_=0)

@operator
class gather:
    aout='y'
    ain ='x', 'layout'

    def apl(node, x, layout):
        return dict(y=layout.gather(x))

    def vjp(engine, _y, layout):
        _x = layout.exchange(_y)
        return dict(_layout=0, _x=_x)

    def jvp(engine, layout, layout_, x_,):
        return dict(y_=layout.gather(x_))

@operator
class exchange:
    aout='y'
    ain ='x', 'layout'

    def apl(node, x, layout):
        return dict(y=layout.exchange(x))

    def vjp(engine, _y, layout):
        _x = layout.gather(_y)
        return dict(_layout=0, _x=_x)

    def jvp(engine, layout, layout_, x_,):
        return dict(y_=layout.exchange(x_))

def fourier_space_neg_gradient(dir, pm):
    nyquist =  numpy.pi / pm.BoxSize[dir] * (pm.Nmesh[dir] - 0.01)
    def filter(k):
        mask = abs(k[dir]) < nyquist
        return - 1j * k[dir] * mask
    return filter

def fourier_space_laplace(k):
    k2 = k.normp(2)
    bad = k2 == 0
    k2[bad] = 1
    k2 = - 1 / k2
    k2[bad] = 0
    return k2

@autooperator('rhok,q->dx1')
def lpt1(rhok, q, pm):
    p = apply_transfer(rhok, fourier_space_laplace)

    layout = decompose(q, pm)

    r1 = []
    for d in range(pm.ndim):
        dx1_c = apply_transfer(p, fourier_space_neg_gradient(d, pm))
        dx1_r = c2r(dx1_c)
        dx1 = readout(dx1_r, q, layout)
        r1.append(dx1)

    dx1 = linalg.stack(r1, axis=-1)

    return dict(dx1 = dx1)

@autooperator('rhok->rho_lpt2')
def lpt2src(rhok, pm):
    if pm.ndim != 3:
        raise ValueError("LPT 2 only works in 3d")

    D1 = [1, 2, 0]
    D2 = [2, 0, 1]

    potk = apply_transfer(rhok, fourier_space_laplace)

    Pii = []
    for d in range(pm.ndim):
        t = apply_transfer(potk, fourier_space_neg_gradient(d, pm))
        Pii1 = apply_transfer(t, fourier_space_neg_gradient(d, pm))
        Pii1 = c2r(Pii1)
        Pii.append(Pii1)

    source = None
    for d in range(pm.ndim):
        source1 = Pii[D1[d]] * Pii[D2[d]]
        if source is None:
            source = source1
        else:
            source = source + source1

    for d in range(pm.ndim):
        t = apply_transfer(potk, fourier_space_neg_gradient(D1[d], pm))
        Pij1 = apply_transfer(t, fourier_space_neg_gradient(D2[d], pm))
        Pij1 = c2r(Pij1)
        source1 = - Pij1 * Pij1
        source = source + source1

    source = (3.0 / 7) * source

    return dict(rho_lpt2=source)

@autooperator('wnk->c')
def induce_correlation(wnk, powerspectrum, pm):
    def tf(k):
        k = sum(ki ** 2 for ki in k) ** 0.5
        return (powerspectrum(k) / pm.BoxSize.prod()) ** 0.5

    c = apply_transfer(wnk, tf)
    return dict(c = c)

@autooperator('rhok->dx1,dx2')
def lpt(rhok, q, pm):

    dx1 = lpt1(rhok, q, pm)
    source2 = lpt2src(rhok, pm)
    rhok2 = r2c(source2)
    dx2 = lpt1(rhok2, q, pm)

    return dict(dx1=dx1, dx2=dx2)


@autooperator('dx->f,potk')
def gravity(dx, q, pm):
    from vmad.core.stdlib import watchpoint
    x = q + dx
    #def w(q): print('q', q)
    #watchpoint(x, w)
    #watchpoint(x, lambda x: print('x', q))
    #watchpoint(dx, lambda dx: print('dx', dx))
    layout = decompose(x, pm)

    xl = exchange(x, layout)
    rho = paint(xl, 1.0, None, pm)

    # convert to 1 + delta
    fac = 1.0 * pm.Nmesh.prod() / pm.comm.allreduce(len(q))

    rho = rho * fac
    rhok = r2c(rho)

    p = apply_transfer(rhok, fourier_space_laplace)

    r1 = []
    for d in range(pm.ndim):
        dx1_c = apply_transfer(p, fourier_space_neg_gradient(d, pm))
        dx1_r = c2r(dx1_c)
        dx1l = readout(dx1_r, xl, None)
        dx1 = gather(dx1l, layout)
        r1.append(dx1)

    f = linalg.stack(r1, axis=-1)
    return dict(f=f, potk=p)

def KickFactor(pt, ai, ac, af):
    return 1 / (ac ** 2 * pt.E(ac)) * (pt.Gf(af) - pt.Gf(ai)) / pt.gf(ac)

def DriftFactor(pt, ai, ac, af):
    return 1 / (ac ** 3 * pt.E(ac)) * (pt.Gp(af) - pt.Gp(ai)) / pt.gp(ac)

@autooperator('dx_i,p_i->dx,p,f')
def leapfrog(dx_i, p_i, q, stages, pt, pm):

    Om0 = pt.Om(1.0)

    dx = dx_i
    p = p_i
    f, potk = gravity(dx, q, pm)

    for ai, af in zip(stages[:-1], stages[1:]):
        ac = (ai * af) ** 0.5

        # kick
        dp = f * (KickFactor(pt, ai, ai, ac) * 1.5 * Om0)
        p = p + dp

        # drift
        ddx = p * DriftFactor(pt, ai, ac, af)
        dx = dx + ddx

        # force
        f, potk = gravity(dx, q, pm)

        # kick
        dp = f * (KickFactor(pt, ac, af, af) * 1.5 * Om0)
        p = p + dp

    f = f * (1.5 * Om0)
    return dict(dx=dx, p=p, f=f)

@autooperator('rhok->dx,p')
def firststep(rhok, q, a, pt, pm):

    dx1, dx2 = lpt(rhok, q, pm)

    E = pt.E(a)
    D1 = pt.D1(a)
    D2 = pt.D2(a)
    f1 = pt.f1(a)
    f2 = pt.f2(a)

    dx1 = dx1 * D1
    dx2 = dx2 * D2

    p1 = dx1 * (a ** 2 * f1 * E)
    p2 = dx2 * (a ** 2 * f2 * E)

    p = p1 + p2
    dx = dx1 + dx2
    return dict(dx=dx, p=p)

@autooperator('rhok->dx,p,f')
def nbody(rhok, q, stages, cosmology, pm):
    from fastpm.background import PerturbationGrowth

    stages = numpy.array(stages)
    mid = (stages[1:] * stages[:-1]) ** 0.5
    support = numpy.concatenate([mid, stages])
    support.sort()
    pt = PerturbationGrowth(cosmology, a=support)

    dx, p = firststep(rhok, q, stages[0], pt, pm)

    dx, p, f = leapfrog(dx, p, q, stages, pt, pm)

    return dict(dx=dx, p=p, f=f)

@operator
class cdot:
    ain = {'x1' : 'ComplexField', 'x2' : 'Complexfield'}
    aout = {'y' : '*'}

    # only keep the real part, assuming two fields are hermitian.
    def apl(node, x1, x2):
        return dict(y=x1.cdot(x2).real)

    def vjp(node, x1, x2, _y):
        _x1 = x2.cdot_vjp(_y)
        _x2 = x1.cdot_vjp(_y)
        return dict(_x1=_x1, _x2=_x2)

    def jvp(node, x1_, x2_, x1, x2):
        return dict(y_=x1.cdot(x2_).real + x2.cdot(x1_).real)

class FastPMSimulation:
    def __init__(self, stages, cosmology, pm, q=None):
        from fastpm.background import PerturbationGrowth

        if q is None:
            q = pm.generate_uniform_particle_grid()

        stages = numpy.array(stages)
        mid = (stages[1:] * stages[:-1]) ** 0.5
        support = numpy.concatenate([mid, stages])
        support.sort()
        pt = PerturbationGrowth(cosmology, a=support)

        self.stages = stages
        self.pt = pt
        self.pm = pm
        self.q = q

    def KickFactor(self, ai, ac, af):
        pt = self.pt
        return 1 / (ac ** 2 * pt.E(ac)) * (pt.Gf(af) - pt.Gf(ai)) / pt.gf(ac)

    def DriftFactor(self, ai, ac, af):
        pt = self.pt
        return 1 / (ac ** 3 * pt.E(ac)) * (pt.Gp(af) - pt.Gp(ai)) / pt.gp(ac)

    @autooperator('rhok->dx, p')
    def firststep(self, rhok):
        q = self.q
        pm = self.pm
        pt = self.pt
        a = self.stages[0]

        dx1, dx2 = lpt(rhok, q, pm)

        E = pt.E(a)
        D1 = pt.D1(a)
        D2 = pt.D2(a)
        f1 = pt.f1(a)
        f2 = pt.f2(a)

        dx1 = dx1 * D1
        dx2 = dx2 * D2

        p1 = dx1 * (a ** 2 * f1 * E)
        p2 = dx2 * (a ** 2 * f2 * E)

        p = p1 + p2
        dx = dx1 + dx2
        return dict(dx=dx, p=p)

    @autooperator('dx->f,potk')
    def gravity(self, dx):
        q = self.q
        pm = self.pm

        x = q + dx

        layout = decompose(x, pm)

        xl = exchange(x, layout)
        rho = paint(xl, 1.0, None, pm)

        # convert to 1 + delta
        fac = 1.0 * pm.Nmesh.prod() / pm.comm.allreduce(len(q))

        rho = rho * fac
        rhok = r2c(rho)

        p = apply_transfer(rhok, fourier_space_laplace)

        r1 = []
        for d in range(pm.ndim):
            dx1_c = apply_transfer(p, fourier_space_neg_gradient(d, pm))
            dx1_r = c2r(dx1_c)
            dx1l = readout(dx1_r, xl, None)
            dx1 = gather(dx1l, layout)
            r1.append(dx1)

        f = linalg.stack(r1, axis=-1)
        return dict(f=f, potk=p)

    @autooperator('rhok->dx,p,f')
    def run(self, rhok):

        dx, p = self.firststep(rhok)

        pt = self.pt
        pm = self.pm
        stages = self.stages
        q = self.q

        Om0 = pt.Om(1.0)

        f, potk = self.gravity(dx)

        for ai, af in zip(stages[:-1], stages[1:]):
            ac = (ai * af) ** 0.5

            # kick
            dp = f * (self.KickFactor(ai, ai, ac) * 1.5 * Om0)
            p = p + dp

            # drift
            ddx = p * self.DriftFactor(ai, ac, af)
            dx = dx + ddx

            # force
            f, potk = self.gravity(dx)

            # kick
            dp = f * (self.KickFactor(ac, af, af) * 1.5 * Om0)
            p = p + dp

        f = f * (1.5 * Om0)
        return dict(dx=dx, p=p, f=f)


